/*
 * Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  http://aws.amazon.com/apache2.0
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

import 'package:amplify_api/src/graphql/model_mutations_factory.dart';
import 'package:amplify_core/amplify_core.dart';

/// Static helpers to generate mutation `GraphQLRequest` instances from models generated by Amplify codegen.

class ModelMutations {
  ModelMutations._(); // only static methods here, prevent calling constructor

  /// Generates a request to create a model.
  ///
  /// ```dart
  /// Todo todo = Todo(name: 'my first todo', description: 'todo description');
  /// final request = ModelMutations.create(todo);
  /// ```
  static GraphQLRequest<T> create<T extends Model>(T model) {
    return ModelMutationsFactory.instance.create<T>(model);
  }

  /// Generates a request to delete a model.
  ///
  ///
  /// ```dart
  /// final request = ModelMutations.delete(someTodoIDontWant);
  /// ```
  ///
  /// An optional `where` parameter can be supplied as a condition for the deletion to be evaluated on the server.
  static GraphQLRequest<T> delete<T extends Model>(T model,
      {QueryPredicate? where}) {
    return ModelMutationsFactory.instance.delete<T>(model, where: where);
  }

  /// Generates a request to delete a model by ID.
  ///
  /// ```dart
  /// final request = ModelMutations.deleteById(Todo.classType, 'some-todo-id-123');
  /// ```
  ///
  /// An optional `where` parameter can be supplied as a condition for the deletion to be evaluated on the server.
  static GraphQLRequest<T> deleteById<T extends Model>(
      ModelType<T> modelType, String id,
      {QueryPredicate? where}) {
    return ModelMutationsFactory.instance
        .deleteById<T>(modelType, id, where: where);
  }

  /// Generates a request to update a model instance.
  ///
  /// ```dart
  /// final todoWithNewName = someTodo.copyWith(name: 'new name');
  /// final request = ModelMutations.update(todoWithNewName);
  /// ```
  ///
  /// An optional `where` parameter can be supplied as a condition for the update to be evaluated on the server.
  static GraphQLRequest<T> update<T extends Model>(T model,
      {QueryPredicate? where}) {
    return ModelMutationsFactory.instance.update<T>(model, where: where);
  }
}
